#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <xcb/xcb.h>
#include <xcb/xcb_event.h>
#include <xcb/xcb_ewmh.h>

#define ARRAY_SIZE(a) (sizeof(a) / sizeof((a)[0]))
#define FONT "-*-terminus-medium-r-*-*-16-*-*-*-*-*-*-*"

static xcb_connection_t *conn;
static xcb_window_t win;
static xcb_gcontext_t foreground;
static uint32_t foreground_color = 0x8f7ad1;
static uint32_t background_color = 0x39106d;
static xcb_font_t font;
static xcb_ewmh_connection_t ewmh;
static int is_mapped = 0;
static int is_visible = 0;

static void
panic(const char *m)
{
	fprintf(stderr, "%s\n", m);
	exit(1);
}

static void
log_error(xcb_generic_error_t *err)
{
	if (err == NULL)
		return;
	fprintf(stderr, "error (request: %s): %s\n",
		xcb_event_get_request_label(err->major_code),
		xcb_event_get_error_label(err->error_code));
}

static const char *
connection_error_message(int err)
{
	switch (err) {
	case XCB_CONN_ERROR:			return "IO error";
	case XCB_CONN_CLOSED_EXT_NOTSUPPORTED:	return "extension not supported";
	case XCB_CONN_CLOSED_MEM_INSUFFICIENT:	return "insufficient memory";
	case XCB_CONN_CLOSED_REQ_LEN_EXCEED:	return "request length exceeded";
	case XCB_CONN_CLOSED_PARSE_ERR:		return "parse error";
	}
	return "unknown";
}

static void
open_connection(void)
{
	int err;
	conn = xcb_connect(NULL, NULL);
	err = xcb_connection_has_error(conn);
	if (err)
		panic(connection_error_message(err));
}

static void
ewmh_init(void)
{
	xcb_intern_atom_cookie_t *cookies;
	xcb_generic_error_t *err;
	cookies = xcb_ewmh_init_atoms(conn, &ewmh);
	if (!xcb_ewmh_init_atoms_replies(&ewmh, cookies, &err)) {
		log_error(err);
		panic("failed to init EWMH");
	}
}

static void
open_font(void)
{
	font = xcb_generate_id(conn);
	xcb_open_font(conn, font, sizeof(FONT)-1, FONT);
}

static void
create_foreground(void)
{
	uint32_t mask =
		XCB_GC_FOREGROUND | XCB_GC_BACKGROUND | XCB_GC_FONT |
		XCB_GC_GRAPHICS_EXPOSURES;
	uint32_t values[] = {
		foreground_color,
		background_color,
		font,
		0,
	};
	foreground = xcb_generate_id(conn);
	xcb_create_gc(conn, foreground, win, mask, values);
}

#define EVENT_MASK \
	XCB_EVENT_MASK_VISIBILITY_CHANGE \
	| XCB_EVENT_MASK_STRUCTURE_NOTIFY

static void
create_window(void)
{
	const xcb_setup_t *setup;
	xcb_screen_t *screen;
	uint32_t mask = XCB_CW_BACK_PIXEL | XCB_CW_EVENT_MASK;
	uint32_t values[] = { background_color, EVENT_MASK };
	static const char name[] = "test5";
	setup = xcb_get_setup(conn);
	if (setup->roots_len < 1)
		panic("no roots");
	screen = xcb_setup_roots_iterator(setup).data;
	win = xcb_generate_id(conn);
	xcb_create_window(
		conn, XCB_COPY_FROM_PARENT, win, screen->root,
		0, 0, 800, 640, 10,
		XCB_WINDOW_CLASS_INPUT_OUTPUT,
		screen->root_visual, mask, values);
	xcb_change_property(conn, XCB_PROP_MODE_REPLACE, win,
		XCB_ATOM_WM_NAME, XCB_ATOM_STRING, 8,
		sizeof(name)-1, name);
	xcb_change_property(conn, XCB_PROP_MODE_REPLACE, win,
		ewmh._NET_WM_NAME, XCB_ATOM_STRING, 8,
		sizeof(name)-1, name);
	xcb_map_window(conn, win);
	if (xcb_flush(conn) <= 0)
		panic("failed to flush buffer");
}

struct sparkline {
	const char *name;
	int data[10];
};

#define MAX_SPARKLINES 32
static struct sparkline sparklines[MAX_SPARKLINES];
static int nr_sparklines;

static void
load_sparkline(const char *name)
{
	static const char *dirname = "./test5-data/";
	char *fullname;
	FILE *file;
	struct sparkline *s;
	int i;
	if (nr_sparklines >= MAX_SPARKLINES)
		return;
	fullname = malloc(sizeof(dirname) + strlen(name));
	strcpy(fullname, dirname);
	strcat(fullname, name);
	file = fopen(fullname, "r");
	s = &sparklines[nr_sparklines];
	s->name = name;
	for (i = 0; i < 10; i++) {
		fscanf(file, "%d", &s->data[i]);
	}
	nr_sparklines++;
}

static void
load_data(void)
{
	load_sparkline("wins");
	load_sparkline("losses");
}

static void
init(void)
{
	open_connection();
	ewmh_init();
	create_window();
	open_font();
	create_foreground();
	load_data();
}

static void
handle_visibility_notify_event(xcb_visibility_notify_event_t *e)
{
	if (e->state == XCB_VISIBILITY_FULLY_OBSCURED)
		is_visible = 0;
	else
		is_visible = 1;
}

static void
handle_map_notify_event(xcb_map_notify_event_t *e)
{
	is_mapped = 1;
}

static void
handle_unmap_notify_event(xcb_unmap_notify_event_t *e)
{
	is_mapped = 0;
}

static void
show_sparkline(int line_nr, const struct sparkline *s)
{
	static char buf[64];
	int len, i;
	xcb_rectangle_t *r, rects[10];
	strcpy(buf, s->name);
	strcat(buf, ":");
	len = strlen(buf);
	memset(buf + len, ' ', ARRAY_SIZE(buf) - len);
	xcb_image_text_8(conn, 24, win, foreground,
		10, 16 * line_nr, buf);
	for (i = 0; i < 10; i++) {
		r = &rects[i];
		r->x = 80 + 4 * i;
		r->y = 16 * line_nr - 2 * (s->data[i] + 1);
		r->width = 2;
		r->height = 2 * (s->data[i] + 1);
	}
	xcb_poly_fill_rectangle(conn, win, foreground, 10, rects);
}

static void
draw(void)
{
	int i, line_nr;
	for (i = 0, line_nr = 1; i < nr_sparklines; i++, line_nr++)
		show_sparkline(line_nr, &sparklines[i]);
	if (xcb_flush(conn) <= 0)
		panic("failed to flush buffer");
}

static void
handle_event(xcb_generic_event_t *e)
{
	int type = XCB_EVENT_RESPONSE_TYPE(e);
	switch (type) {
	case XCB_VISIBILITY_NOTIFY:
		handle_visibility_notify_event(
			(xcb_visibility_notify_event_t *)e);
		break;
	case XCB_MAP_NOTIFY:
		handle_map_notify_event((xcb_map_notify_event_t *)e);
		break;
	case XCB_UNMAP_NOTIFY:
		handle_unmap_notify_event((xcb_unmap_notify_event_t *)e);
		break;
	}
	if (is_mapped && is_visible)
		draw();
}

static void
handle_error(xcb_generic_error_t *err)
{
	log_error(err);
}

static void
interact(void)
{
	xcb_generic_event_t *e;
	for (;;) {
		e = xcb_wait_for_event(conn);
		if (e == NULL) {
			int err;
			err = xcb_connection_has_error(conn);
			if (err)
				panic(connection_error_message(err));
			panic("unexpected NULL event");
		}
		if (e->response_type)
			handle_event(e);
		else
			handle_error((xcb_generic_error_t *)e);
		free(e);
	}
}

int
main(void)
{
	init();
	interact();
	return 0;
}
